import argparse
import logging
from getpass import getpass
from sqlalchemy import or_
from broker_clients import session
from broker_clients import Accounts
from binance.client import Client
from broker_clients.quantfury_proxy.broker.client import Quantfury


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class BinanceAuth:
    @classmethod
    def are_valid_tokens(cls, api_key, secret_key):
        client = Client(api_key=api_key, api_secret=secret_key, tld='vision')
        account = client.get_account()
        if 'canWithdraw' not in account:
            logger.info('func=Accounts.are_valid_tokens, msg=Invalid tokens')
            return False
        elif account.get('canWithdraw'):
            logger.info('func=Accounts.are_valid_tokens, msg=Excessive permissions')
            return False
        else:
            return True

    @classmethod
    def create(cls, name, api_key, secret_key):
        account = session.query(Accounts).filter(or_(
            Accounts.name == name, Accounts.api_key == api_key, Accounts.secret_key == secret_key)).first()

        if not cls.are_valid_tokens(api_key, secret_key):
            return False
        if account:
            logger.info(
                'func=Accounts.create_account, msg=Balance (name api_key secret_key) combination already exists')
            return False
        account = Accounts(name=name, api_key=api_key, secret_key=secret_key, broker_name='binance')
        session.add(account)
        session.commit()
        if not cls.are_valid_tokens(account.api_key, account.secret_key):
            logger.error('func=Accounts.create_account, msg=Cipher error')
            return False

        return True


class QuantfuryAuth:
    @classmethod
    def generate_new_token(cls, api_key, secret_key):

        login_result = Quantfury.login(api_key, secret_key)
        if 'data' in login_result and 'token' in login_result['data']:
            return login_result['data']['token']
        else:
            logger.info('func=Accounts.are_valid_tokens, msg=Invalid tokens')
            return False

    @classmethod
    def create(cls, name, api_key, secret_key):
        account = session.query(Accounts).filter(or_(
            Accounts.name == name, Accounts.api_key == api_key)).first()

        token = cls.generate_new_token(api_key, secret_key)
        if not token:
            logger.error('func=Accounts.create_account, msg=Invalid Token')
            return False
        if account and token:
            account.token = token
            logger.info(
                'func=Accounts.create_account, msg=token updated')
        else:
            account = Accounts(name=name, api_key=api_key, secret_key=secret_key, token=token, broker_name='quantfury')
            session.add(account)
        session.commit()

        return True

    @classmethod
    def update(cls, id_name):
        if type(id_name) is str:
            account = session.query(Accounts).filter(Accounts.name == id_name).first()
        else:
            account = session.query(Accounts).filter(Accounts.account_id == id_name).first()

        if not account:
            logger.error('func=Accounts.update, account_id_name={}, msg=Invalid account name'.format(id_name))
            raise ValueError('Invalid account name')

        token = cls.generate_new_token(account.api_key, account.secret_key)
        if not token:
            logger.error('func=Accounts.update, account_id_name={}, msg=Invalid Credentials'.format(id_name))
            raise ValueError('Invalid Credentials')
        account.token = token
        session.commit()

        return True


def parse_args():
    parser = argparse.ArgumentParser(description='Crate account')
    parser.add_argument('--broker', dest='broker', required=True, choices=['binance', 'quantfury'],
                        help='Broker of trading')
    parser.add_argument('--name', dest='name', required=True,
                        help='Account name')
    parser.add_argument('--api_key', dest='api_key', required=True,
                        help='API key or username')

    args = parser.parse_args()

    return args


def command():
    args = parse_args()

    args.secret_key = getpass("secret_key:")
    auth = args.broker
    if auth == 'binance':
        auth = BinanceAuth()
    else:
        auth = QuantfuryAuth()

    if auth.create(args.name, args.api_key, args.secret_key):
        logger.info('func=__main__, msg=Account created')


if __name__ == '__main__':
    command()



from dotenv import load_dotenv
import os
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

load_dotenv()

Base = declarative_base()
BROKER_DB = os.getenv("BROKER_DB")
engine = create_engine(BROKER_DB, echo=False)
Session = sessionmaker(bind=engine)
session = Session()

from broker_clients.emulator_proxy.storage.models import Accounts, Balances, Trades
#from broker_clients.binance_proxy.storage.models import Orders, Fills, SymbolMetadata
#from broker_clients.quantfury_proxy.storage.models import QuantfuryPositionType, QuantfuryPositions, QuantfuryOrders



import sqlite3
import os
from dotenv import load_dotenv

load_dotenv()

BROKER_DATABASE_NAME = os.getenv("BROKER_DATABASE_NAME")
MINIMUM_TRADE_AMOUNT = 0


class Broker:
    def __init__(self, symbol, time_frame):
        self._symbol = symbol
        self._time_frame = time_frame
        self._price = None
        self._stop_loss = None
        self._amount = None

    def clean_order(self):
        self._price = None
        self._stop_loss = None
        self._amount = None

    def market(self, side, amount, stop_loss=None, stop_loss_percentage=None, emulated_time=None):
        pass

    def has_open_order(self):
        return self._price is not None

    def has_earning(self, emulated_time=None):
        return self._price < emulated_time['Low']

    def take_profit(self, row=None):
        pass

    def has_opened_position(self):
        return self.price is not None and self.stop_loss is not None and self.price > self.stop_loss

    @property
    def stop_loss(self):
        return self._stop_loss

    @stop_loss.setter
    def stop_loss(self, value):
        self._stop_loss = value

    @property
    def price(self):
        return self._price

    @price.setter
    def price(self, value):
        self._price = value

    @property
    def amount(self):
        return self._amount

    @amount.setter
    def amount(self, value):
        self._amount = value

    def set_sell_price(self, sell_price):
        pass


class BrokerMemoryEmulator(Broker):

    def market(self, side, amount, stop_loss=None, stop_loss_percentage=None, emulated_time=None):
        if side not in ['BUY', 'SELL'] or amount <= MINIMUM_TRADE_AMOUNT:
            raise ValueError('Invalid side: {} or amount: {} in market order'.format(side, amount))
        if stop_loss is None and stop_loss_percentage is None:
            raise ValueError('Stop loss must be set')

        if side == 'BUY':
            self._price = (emulated_time['Close'] + emulated_time['Open']) / 2
            if not stop_loss:
                self.stop_loss = self.price - (self.price * stop_loss_percentage)
            self._amount = amount

    def take_profit(self, row=None):
        delta = row['Low'] - self._price
        profit_percentage = delta / row['Low']
        profit_amount = self._amount * profit_percentage
        self.clean_order()
        return {
            "profit_percentage": profit_percentage,
            "profit_amount": profit_amount
        }


class BrokerDiskEmulator(Broker):
    def __init__(self, symbol, time_frame, database_name=BROKER_DATABASE_NAME):
        self.conn = sqlite3.connect(database_name)
        super().__init__(symbol, time_frame)
        self._create_table()
        if self.trading_budget() is None:
            self.set_trading_budget(100)

    def _create_table(self):
        self.conn.execute('''CREATE TABLE IF NOT EXISTS trades(
        date text,
        side text,
        symbol text,
        time_frame text,
        amount real,
        stop_loss real,
        buy_price real,
        sell_price real,
        profit real
        )''')

        self.conn.execute('''CREATE TABLE IF NOT EXISTS balances(
                symbol text,
                time_frame text,
                free real,
                locked real
                )''')

    def trading_budget(self):
        result_set = self.conn.execute("""SELECT free FROM balances WHERE symbol=? and time_frame=?""", (self._symbol, self._time_frame))
        row = result_set.fetchone()
        if row:
            return row[0]

    def set_trading_budget(self, value):
        if not self.trading_budget():
            trade = [(self._symbol, self._time_frame, value, 0)]
            self.conn.executemany('INSERT INTO balances VALUES (?,?,?,?)', trade)
        else:
            self.conn.execute("""UPDATE balances
            SET free = ?
            WHERE symbol=? and time_frame=?""", (value, self._symbol, self._time_frame))
        self.conn.commit()

    def init_broker_object(self):
        result_set = self.conn.execute("""
           SELECT buy_price, stop_loss, amount 
           FROM trades WHERE symbol=? and time_frame=? and sell_price is NULL""", (self._symbol, self._time_frame))
        row = result_set.fetchone()
        if row:
            self._price, self._stop_loss, self._amount = row[0], row[1], row[2]

    def market(self, side, amount, stop_loss=None, stop_loss_percentage=None, emulated_time=None):
        if side not in ['BUY', 'SELL'] or amount <= MINIMUM_TRADE_AMOUNT:
            raise ValueError('Invalid side: {} or amount: {} in market order'.format(side, amount))
        if stop_loss is None and stop_loss_percentage is None:
            raise ValueError('Stop loss must be set')

        if side == 'BUY':
            self.price = emulated_time['Close']
            if not stop_loss:
                self.stop_loss = self.price - (self.price * stop_loss_percentage)
            self._amount = amount
            trade = [(emulated_time['date'], side, self._symbol, self._time_frame, amount, self.stop_loss, self.price, None, None)]
            self.conn.executemany('INSERT INTO trades VALUES (?,?,?,?,?,?,?,?,?)', trade)
            self.conn.commit()

    def take_profit(self, row=None):
        current_price = row['Close']
        delta = current_price - self.price
        profit_percentage = delta / current_price
        profit_amount = self._amount * profit_percentage
        self.set_sell_price(current_price, self._amount + profit_amount)
        self.clean_order()
        return {
            "profit_percentage": profit_percentage,
            "profit_amount": profit_amount
        }

    def set_sell_price(self, sell_price, profit):
        self.conn.execute("begin")
        update_query = """UPDATE trades SET sell_price = ?, profit = ? 
        WHERE symbol=? and time_frame=? and sell_price is NULL"""
        self.conn.execute(update_query, (sell_price, profit, self._symbol, self._time_frame))
        self.set_trading_budget(profit)


if __name__ == '__main__':
    broker = BrokerDiskEmulator('BINANCE:BTCUSDT', '60')
    budget = broker.trading_budget()
    print(budget)
    #broker.market('BUY', 100, None, 0.03, {'date': '2020/10/28, 18:00:00', 'Close': 5000})



from decimal import Decimal

from broker_clients.binance_proxy.broker import UtilsAPI
from broker_clients import session
from broker_clients.emulator_proxy.storage.models import Orders, SymbolMetadata
from broker_clients import Accounts, Trades, Balances
from binance.client import Client
import logging

logger = logging.getLogger(__name__)


class ExtendedAccount:

    def __init__(self, name, symbol):
        self.name = name
        account = session.query(Accounts).filter(Accounts.name == name).first()
        self.client = Client(api_key=account.api_key, api_secret=account.secret_key, tld='vision')
        self.symbol = symbol
        self.symbol_metadata = session.query(SymbolMetadata).filter(SymbolMetadata.symbol == self.symbol).first()

    def calculate_precise_quantity(self, quote_order_quantity, price):
        quantity = self.quantity_from_quote(quote_order_quantity, price, self.minimum_quantity_lot_size)
        quantity = "{:0.0{}f}".format(quantity, self.quote_precision)
        return quantity

    def update_buy_price(self, trade_id, order):
        pass

    def market(self, trade_id, symbol, side, quote_order_quantity):
        try:
            order = Orders.register_order(self.client, trade_id, symbol=symbol, side=side,
                                          quoteOrderQty=quote_order_quantity, type=self.client.ORDER_TYPE_MARKET)
            return order
        except Exception as e:
            logger.error('func=ExtendedAccount.market, account_name={}, trade_id={}, symbol={},'
                         'side={}, quote_order_quantity={}, msg={}'.format(
                          self.name, trade_id, symbol, side, quote_order_quantity, e))
            raise e

    def market_quantity(self, trade_id, symbol, side, quantity):
        try:
            order = Orders.register_order(self.client, trade_id, symbol=symbol, side=side,
                                          quantity=quantity, type=self.client.ORDER_TYPE_MARKET)
            return order
        except Exception as e:
            logger.error('func=ExtendedAccount.market, account_name={}, trade_id={}, symbol={},'
                         'side={}, quantity={}, msg={}'.format(
                          self.name, trade_id, symbol, side, quantity, e))
            raise e

    def limit_buy(self, trade_id, symbol, quote_order_quantity, price):
        try:
            quantity = self.calculate_precise_quantity(quote_order_quantity, price)
            res = Orders.register_order(self.client, trade_id, symbol=symbol, quantity=quantity, price=price,
                                        side=self.client.SIDE_BUY, type=self.client.ORDER_TYPE_LIMIT)
            return res
        except Exception as e:
            logger.error('func=ExtendedAccount.limit_buy, account_name={}, trade_id={}, symbol={},'
                         'side={}, quote_order_quantity={}, msg={}'.format(
                          self.name, trade_id, symbol, side, quote_order_quantity, e))
            raise e

    def limit_sell(self, trade_id, symbol, quote_order_quantity, price):
        try:
            quantity = self.calculate_precise_quantity(quote_order_quantity, price)

            res = Orders.register_order(self.client, trade_id, symbol=symbol, quantity=quantity,
                                        price=price, side=self.client.SIDE_SELL, type=self.client.ORDER_TYPE_LIMIT)
            return res
        except Exception as e:
            logger.error('func=ExtendedAccount.limit_sell, account_name={}, trade_id={}, symbol={},'
                         'side={}, quote_order_quantity={}, msg={}'.format(
                          self.name, trade_id, symbol, side, quote_order_quantity, e))
            raise e

    def update_limit_orders(self):
        records = session.query(Orders.order_id, Orders.symbol).filter(Orders.status != 'FILLED').all()
        for row in records:
            order_id, symbol = row[0], row[1]
            order_response = self.client.get_order(symbol=symbol, orderId=order_id)
            Orders.update_order(order_response)

    def take_profit_limit(self, symbol, side, quote_order_quantity, price, stop_price):
        try:
            quantity = self.calculate_precise_quantity(quote_order_quantity, price)

            order_response = self.client.create_order(symbol=symbol, side=side, quantity=quantity, price=price,
                                                      stopPrice=stop_price, type=self.client.ORDER_TYPE_STOP_LOSS_LIMIT,
                                                      timeInForce=self.client.TIME_IN_FORCE_GTC)
            res = Orders.register_order(order_response)
            return res

        except Exception as e:
            logger.error('func=ExtendedAccount.take_profit_limit, symbol={}, quote_order_quantity={}, price={}, '
                         'msg={}'.format(symbol, quote_order_quantity, price, e))
            raise e

    def test_order(self, symbol, side, quote_order_quantity):
        klines = self.client.get_historical_klines(symbol, self.client.KLINE_INTERVAL_15MINUTE, "1 hour ago UTC")
        close_price = Decimal(klines[0][4])
        # a = limit_buy(symbol, quote_order_quantity, close_price)

        limit_price = close_price * Decimal('0.97')
        stop_price = close_price - Decimal('0.96')
        b = self.stop_loss_limit(symbol, side, quote_order_quantity, limit_price, stop_price)

    def fill_empty_prices(self, symbol):
        filled_orders = session.query(Orders).filter_by(status=self.client.ORDER_STATUS_FILLED, price=0)
        zero_price_trades = session.query(Orders).join(Trades).filter(Trades.buy_price == 0)
        orders = list(filled_orders) + list(zero_price_trades)

        my_trades = self.client.get_my_trades(symbol=symbol)
        for order in orders:
            my_trade = self.find_by_order_id(my_trades, order.order_id)
            if my_trade:
                order.fill_empty_price(UtilsAPI.clean_response(my_trade))
            else:
                logger.critical('func=ExtendedAccount.fill_empty_prices, order_id={}, symbol={}, '
                                'msg=No trade found'.format(order.order_id, symbol))

    @classmethod
    def find_by_order_id(cls, my_trades, order_id):
        for t in my_trades:
            if t['orderId'] == order_id:
                return t

    def update_symbol_metadata(self):
        symbols = session.query(Balances.symbol).distinct(Balances.symbol)
        for row in symbols:
            symbol_text = row[0].split(':')[1]
            symbol_metadata = session.query(SymbolMetadata).filter_by(symbol=symbol_text).first()

            info = self.client.get_symbol_info(symbol_text)
            minimum_quantity_lot_size = Decimal(
                [i for i in info['filters'] if 'LOT_SIZE' == i['filterType']][0]['minQty'])
            minimum_price_filter = Decimal(
                [i for i in info['filters'] if 'PRICE_FILTER' == i['filterType']][0]['minPrice'])
            minimum_notional = Decimal(
                [i for i in info['filters'] if 'MIN_NOTIONAL' == i['filterType']][0]['minNotional'])
            quote_precision = info['quotePrecision']

            if not symbol_metadata:
                symbol_metadata = SymbolMetadata(
                    symbol=symbol_text, minimum_quantity_lot_size=minimum_quantity_lot_size,
                    quote_precision=quote_precision, minimum_price_filter=minimum_price_filter,
                    minimum_notional=minimum_notional)
                session.add(symbol_metadata)
                session.commit()
                self.symbol_metadata = symbol_metadata
            else:
                symbol_metadata.minimum_quantity_lot_size = minimum_quantity_lot_size
                symbol_metadata.minimum_price_filter = minimum_price_filter
                symbol_metadata.minimum_notional = minimum_notional
                symbol_metadata.minimum_notional = minimum_notional
                symbol_metadata.quote_precision = quote_precision
                session.commit()

    def market_with_stop_loss(self, balance_name, side):
        stage = 'API.read_price'
        try:
            ############################################################################################################
            # Block of code only for test net
            # In real escenario Trades tables records are created by the strategy
            self.update_symbol_metadata()
            klines = self.client.get_historical_klines(
                self.symbol, self.client.KLINE_INTERVAL_15MINUTE, "1 hour ago UTC")
            close_price = Decimal(klines[0][4])

            stage = 'SQL.create_trade'
            balance = session.query(Balances).filter_by(name=balance_name).first()
            quantity = self.symbol_metadata.quantity_from_quote(balance.free, close_price)

            trade = Trades(name=balance.name,
                           symbol=balance.symbol,
                           time_frame=balance.time_frame,
                           entry_date='2021-03-27 21:45:00.000000',
                           amount=balance.free,
                           quantity=quantity,
                           stop_loss=close_price*Decimal('0.97'),
                           buy_price=close_price,
                           sell_price=None,
                           profit=None,
                           side='Buy',
                           low_moment_price=close_price,
                           balance_id=balance.balance_id)
            session.add(trade)
            session.commit()
            ############################################################################################################
            stage = 'ExtendedAccount.create_order'

            self.market_quantity(
                trade.trade_id, self.symbol, side, self.symbol_metadata.fix_price_precision(trade.quantity))

            stage = 'API.read_entry_price'
            self.fill_empty_prices(self.symbol)

            stage = 'ExtendedAccount.register_order'

            stop = self.symbol_metadata.calculate_stop_from_limit(trade.stop_loss)
            Orders.register_order(
                self.client, trade.trade_id, symbol=symbol, side=self.client.SIDE_SELL, quantity=trade.quantity,
                price=self.symbol_metadata.fix_price_precision(trade.stop_loss),
                timeInForce=self.client.TIME_IN_FORCE_GTC, stopPrice=self.symbol_metadata.fix_price_precision(stop),
                type=self.client.ORDER_TYPE_STOP_LOSS_LIMIT)

            print(trade)
            print(trade.orders)
        except Exception as e:
            logger.critical('func=ExtendedAccount.market_with_stop_loss, account_name={}, symbol={}, stage={}, '
                            'msg={}'.format(self.name, self.symbol, stage, e))


if __name__ == '__main__':
    symbol = 'BTCUSDT'
    side = 'BUY'
    #quote_order_quantity = Decimal(100)
    balance_name = 'BTC_1D_no_trend_4_testnet'
    balance = session.query(Balances).filter_by(name=balance_name).first()
    if not balance:
        balance = Balances(
                 account_id=1,
                    name=balance_name,
                    symbol='BINANCE:BTCUSDT',
                    time_frame='1D',
                    free=500,
                    locked=0,
                    stop_loss_percentage=0.04,
                    profit_percentage=0.04,
                    trade_in_trend='none',
                    second_scale_from_database=False
        )
        session.add(balance)
        session.commit()
    account = ExtendedAccount('testnet', symbol)
    account.market_with_stop_loss(balance_name, side)









import enum
import logging
from decimal import Decimal
from datetime import datetime
from dotenv import load_dotenv
from sqlalchemy.orm import relationship, validates
from sqlalchemy import Column, Integer, String, Numeric, ForeignKey, DateTime, Enum
from broker_clients import Base, session
from broker_clients.binance_proxy.broker import UtilsAPI
from broker_clients.emulator_proxy.storage.models import SymbolMetadata
from broker_clients.quantfury_proxy.broker.client import Utils, Quantfury
from broker_clients.quantfury_proxy.broker.client import REVERTED_POSITION_TYPES, REVERTED_ORDER_TYPES


load_dotenv()
logger = logging.getLogger(__name__)


class QuantfuryOrderStatus(enum.Enum):
    hold = 'hold'
    open = 'open'
    close = 'close'
    cancel = 'cancel'
    target = 'target'


class QuantfuryOrderType(enum.Enum):
    stop_order = 'stop_order'
    target_order = 'target_order'
    limit_order = 'limit_order'


class QuantfuryPositionType(enum.Enum):
    long = 'long'
    short = 'short'


class QuantfuryPositions(Base):
    __tablename__ = 'quantfury_positions'

    quantfury_order_id = Column(Integer, primary_key=True, nullable=False)
    trade_id = Column(Integer, ForeignKey('trades.trade_id'), nullable=False)

    symbol = Column(String, nullable=False)

    limit_id = Column(String, nullable=False, unique=True)
    limit_price = Column(Numeric, nullable=True)
    limit_quantity = Column(Numeric, nullable=False)
    limit_amount_instrument = Column(Numeric, nullable=False)
    limit_amount_system = Column(Numeric, nullable=False)
    limit_stop_order = Column(Numeric, nullable=False)
    limit_create_date = Column(DateTime, nullable=True)
    limit_execution_type = Column(Numeric, nullable=True)

    api_position_id = Column(String, nullable=True, unique=True)
    position_open_price = Column(Numeric, nullable=True)
    position_quantity = Column(Numeric, nullable=True)
    position_amount_instrument = Column(Numeric, nullable=True)
    position_amount_system = Column(Numeric, nullable=True)
    position_open_date = Column(DateTime, nullable=True)
    position_order = Column(Integer, nullable=True)
    position_session_currency = Column(String, nullable=True)
    position_session_id = Column(String, nullable=True)
    position_spread_adjustment_end_date = Column(DateTime, nullable=True)

    profit_and_loss_instrument = Column(Numeric, nullable=True)
    profit_and_loss_system = Column(Numeric, nullable=True)
    profit_and_loss_account = Column(Numeric, nullable=True)
    close_price = Column(Numeric, nullable=True)
    close_date = Column(DateTime, nullable=True)

    #_position_type = Column("position_type", Enum(QuantfuryPositionType), nullable=True)
    position_type = Column(Enum(QuantfuryPositionType), nullable=True)
    status = Column(Enum(QuantfuryOrderStatus), nullable=False, default=False)  # hold, open, close, cancel

    trade = relationship("Trades", back_populates="quantfury_position", uselist=False)
    quantfury_orders = relationship("QuantfuryOrders", back_populates="position", uselist=True)

    @property
    def stop_order(self):
        order = session.query(QuantfuryOrders).filter_by(
            api_position_id=self.api_position_id, order_type=QuantfuryOrderType.stop_order).first()
        return order

    @property
    def target_order(self):
        order = session.query(QuantfuryOrders).filter_by(
            api_position_id=self.api_position_id, order_type=QuantfuryOrderType.target_order).first()
        return order

    @validates('limit_create_date')
    def validate_limit_create_date(self, key, limit_create_date):
        return datetime.fromtimestamp(limit_create_date)

    @validates('position_spread_adjustment_end_date')
    def validate_position_spread_adjustment_end_date(self, key, position_spread_adjustment_end_date):
        return datetime.fromtimestamp(position_spread_adjustment_end_date)

    @validates('position_open_date')
    def validate_position_open_date(self, key, position_open_date):
        return datetime.fromtimestamp(position_open_date)

    @validates('close_date')
    def validate_close_date(self, key, close_date):
        return datetime.fromtimestamp(close_date)

    def __repr__(self):
        return "<QuantfuryPositions(symbol={}, limit_quantity={}, limit_price={}, position_type={})>".format(
            self.symbol, self.limit_quantity, self.limit_price, self.position_type)

    def update(self, data):
        key_errors = []

        def set_value(key, value):
            try:
                if not hasattr(self, key):
                    raise AttributeError(key)
                setattr(self, key, value)
            except Exception as e:
                key_errors.append(e.__str__())

        for key, value in data.items():
            set_value(key, value)

        if key_errors:
            raise KeyError(';'.join(key_errors))
        try:
            session.commit()
        except Exception as e:
            logger.info('func=QuantfuryOrders.update, msg={}'.format(e))

    def cancel(self, client):
        self.cancel_limit_order(client, self.limit_id)

    def close_with_earnings(self, client, target_price):
        stage, code = 'self.fix_decimals', None
        try:
            # TODO un hardcode target price quotes
            symbol_metadata = session.query(SymbolMetadata).filter(
                SymbolMetadata.currency_pair == self.symbol).first()
            if symbol_metadata:
                limit_amount_system = symbol_metadata.quote_quantity(self.limit_amount_system)
                upper_target_price = symbol_metadata.fix_price_precision(target_price+50)
                lower_target_price = symbol_metadata.fix_price_precision(target_price-50)
            stage = 'API.update_target_position'

            api_order_id = client.active_order_in_position(self.symbol, QuantfuryOrderType.target_order.value)
            if api_order_id:
                target_json_result = client.update_position(
                    api_order_id, self.symbol, upper_target_price,
                    limit_amount_system, QuantfuryOrderType.target_order.value)
            else:
                target_json_result = client.create_position(
                    self.api_position_id, self.symbol, upper_target_price,
                    limit_amount_system, QuantfuryOrderType.target_order.value)

            code = target_json_result.get('code', None)
            if code != 'Success':
                raise ConnectionError('First Update failed')

            stage = 'API.update_stop_loss_position'
            api_order_id = client.active_order_in_position(self.symbol, QuantfuryOrderType.stop_order.value)
            if api_order_id:
                stop_json_result = client.update_position(
                    api_order_id, self.symbol, lower_target_price,
                    limit_amount_system, QuantfuryOrderType.stop_order.value)
                code = stop_json_result.get('code', None)
            else:
                stop_json_result = client.create_position(
                    self.api_position_id, self.symbol, lower_target_price,
                    limit_amount_system, QuantfuryOrderType.stop_order.value)
                code = stop_json_result.get('code', None)

            if code != 'Success':
                raise ConnectionError('Second Update failed')
            stage = 'SQL.update_quantfury_positions'
            self.status = QuantfuryOrderStatus.target
            session.commit()
            logger.info('func=QuantfuryPositions.close_with_earnings, trade_id={}, balance_id={}, '
                        'lower_target_price={}, upper_target_price={}'.format(
                            self.trade.trade_id, self.trade.balance.balance_id, target_price-50, target_price+50))
        except Exception as e:
            logger.critical('func=QuantfuryPositions.close_with_earnings, trade_id={}, balance_id={}, '
                            'target_price={}, stage={}, code={}, msg={}'.format(
                                self.trade.trade_id, self.trade.balance.balance_id, target_price, stage, code, e))

    def update_limit_order(self, client, stop_order=None, limit_price=None):
        stage = 'API.update_limit_order'
        try:
            response = client.update_order(
                self.limit_id, self.symbol, limit_price or self.limit_price, stop_order or self.limit_stop_order,
                None, self.limit_quantity, self.position_type)
            response = UtilsAPI.convert_keys_to_snake_case(response)
            return response
        except Exception as e:
            logger.critical('func=QuantfuryPositions.update_limit_order, limit_id={}, '
                            'stop_order={}, limit_price={}, stage={}, msg={}'.format(
                self.limit_id, stop_order, limit_price, stage, e))

    @classmethod
    def register_order(cls, client, trade_id, **kwargs):
        order_id = None
        stage = 'API.create_order'  # for error control
        try:
            active_trade_id = client.active_trade(kwargs['symbol'])
            if active_trade_id:
                raise ConnectionError('Active trade with id: {}'.format(active_trade_id))
            symbol_metadata = session.query(SymbolMetadata).filter(
                SymbolMetadata.currency_pair == kwargs['symbol']).first()
            if symbol_metadata:
                kwargs['quote_order_quantity'] = symbol_metadata.quote_quantity(kwargs['quote_order_quantity'])
                kwargs['buy_price'] = symbol_metadata.fix_price_precision(kwargs['buy_price'])
                if 'stop_price' in kwargs:
                    kwargs['stop_price'] = symbol_metadata.fix_price_precision(kwargs['stop_price'])
                if 'target_price' in kwargs:
                    kwargs['target_price'] = symbol_metadata.fix_price_precision(kwargs['target_price'])
            else:
                logger.warning('func=QuantfuryOrders.register_order, order_id={}, trade_id={}, '
                               'stage={}, msg=No able to round price or quantity'.format(order_id, trade_id, stage))

            create_order_response = client.create_OCO_long_order(**kwargs)
            if 'data' not in create_order_response and create_order_response.get('code') != 'Success':
                raise ConnectionError(create_order_response.get('code', 'Empty Response'))
            clean_create_order_response = create_order_response['data']
            clean_create_order_response = UtilsAPI.convert_keys_to_snake_case(clean_create_order_response)
            order_id = clean_create_order_response.get('id')

            clean_create_order_response = QuantfuryPositions.rename_fields_from_response(clean_create_order_response)
            clean_create_order_response['trade_id'] = trade_id
            clean_create_order_response['status'] = 'hold'
            clean_create_order_response['position_type'] = REVERTED_POSITION_TYPES[
                clean_create_order_response['position_type']]
            clean_create_order_response['symbol'] = clean_create_order_response.pop('limit_short_name')

            stage = 'SQL.create_order'
            order = QuantfuryPositions(**clean_create_order_response)
            session.add(order)
            session.commit()
            logger.info('func=QuantfuryOrders.register_order, trade_id={}, balance_id={}, '
                        'limit_price={}, limit_stop_order={}'.format(
                order.trade.trade_id, order.trade.balance.balance_id, order.limit_price, order.limit_stop_order))
            return order
        except Exception as e:
            logger.error('func=QuantfuryOrders.register_order, order_id={}, trade_id={}, stage={}, msg={}'.format(
                order_id, trade_id, stage, e))

            raise e

    @classmethod
    def is_equal_price(cls, price_1, price_2, symbol):
        symbol_metadata = session.query(SymbolMetadata).filter_by(currency_pair=symbol).first()
        return abs(price_1 - price_2) <= Decimal(symbol_metadata.minimum_price_filter) * 250

    @classmethod
    def match_limit_with_position(cls, limit, open_positions):
        for position in open_positions:
            position['open_price'] = Decimal(str(position['open_price']))
            if limit.symbol == position['short_name'] and limit.limit_amount_instrument == position['amount_instrument']:
                if len(position.get('stop_orders', [])) > 0 and limit.limit_stop_order is not None:
                    if cls.is_equal_price(Decimal(position['stop_orders'][0]['price']), limit.limit_stop_order, limit.symbol) and \
                            position['stop_orders'][0]['amount_instrument'] == limit.limit_amount_instrument:
                        return position
                else:
                    logger.warning(
                        'func=QuantfuryOrders.match_limit_with_position, symbol={}, limit_price={}, '
                        'limit_stop_order={}, msg=Found position but without stop loss'.format(
                            limit.symbol, limit.limit_price, limit.limit_stop_order))
                    return position
            logger.error(
                'func=QuantfuryOrders.match_limit_with_position, symbol={}, limit_price={}, limit_stop_order={}, '
                'msg=No position Found'.format(limit.symbol, limit.limit_price, limit.limit_stop_order))

    @classmethod
    def is_there_mismatches(cls, open_positions):
        null_position = session.query(QuantfuryOrders).filter_by(position_id=None).all()

        position_ids = [i['id'] for i in open_positions]

        orders_with_positions = session.query(
            QuantfuryOrders).filter(QuantfuryOrders.position_id.in_(position_ids)).all()

        unmatched_position = set(position_ids).difference(set([i.position_id for i in orders_with_positions]))
        unmatched_orders = set(set([i.position_id for i in orders_with_positions])).difference(position_ids)

        if unmatched_orders:
            logger.critical(
                'func=QuantfuryOrders.is_there_mismatches, unmatched_position={}, unmatched_position={}, '
                'msg=Found unmathced orders'.format(';'.join(unmatched_position), ';'.join(unmatched_position)))
            return False

        if unmatched_position:
            logger.critical(
                'func=QuantfuryOrders.is_there_mismatches, unmatched_position={}, unmatched_position={}, '
                'msg=Found unmathced orders'.format(';'.join(unmatched_position), ';'.join(unmatched_position)))
            return False
        return True

    @classmethod
    def sync_orders(cls, client):
        orders = client.list_orders().get('data', [])
        orders = UtilsAPI.convert_keys_to_snake_case(orders)
        orders_with_null_position = session.query(QuantfuryPositions).filter_by(api_position_id=None).all()

        for position in orders_with_null_position:
            new_position = cls.match_limit_with_position(position, orders['open_positions'])
            if new_position:
                stop_orders = new_position.pop('stop_orders')
                #new_position['order_type'] = REVERTED_ORDER_TYPES[new_position.pop('order_type')]
                #stop_orders = UtilsAPI.rename_keys_with_prefix(stop_orders, 'order_')
                stop_orders[0]['api_order_id'] = stop_orders[0].pop('id')
                stop_orders[0]['api_position_id'] = stop_orders[0].pop('trading_position_id')
                stop_orders[0]['order_type'] = REVERTED_ORDER_TYPES[stop_orders[0].pop('order_type')]

                new_position = UtilsAPI.rename_keys_with_prefix(new_position, 'position_')
                new_position['api_position_id'] = new_position.pop('position_id')
                new_position['position_type'] = REVERTED_POSITION_TYPES[new_position.pop('position_position_type')]
                new_position['status'] = QuantfuryOrderStatus.open.value
                del new_position['position_short_name']
                del new_position['position_target_orders']
                position.update(new_position)

                stop_order = QuantfuryOrders(**stop_orders[0])
                session.add(stop_order)
                session.commit()
                logger.info('func=QuantfuryPositions.sync_orders, balance_id={}, trade_id={}, symbol={}, '
                            'position_open_price={}'.format(position.trade.balance_id, position.trade_id,
                                                            position.symbol, new_position['position_open_price']))

    @classmethod
    def sync_closed_positions(cls, client):
        closed_positions = client.list_closed_positions()
        closed_positions = closed_positions.get('data')
        closed_positions = UtilsAPI.convert_keys_to_snake_case(closed_positions)
        closed_positions = [{
            "api_position_id": i.get('id'),
            "close_price": i.get('close_price'),
            "close_date": i.get("close_date"),
            "profit_and_loss_instrument": i.get("pnl_instrument"),
            "profit_and_loss_system": i.get("pnl_system"),
            "profit_and_loss_account": i.get("pnl_account"),
        } for i in closed_positions]
        for close_position in closed_positions:
            stored_position = session.query(QuantfuryPositions).filter_by(
                api_position_id=close_position['api_position_id']).first()
            if stored_position:
                stored_position.update(close_position)
                stored_position.status = QuantfuryOrderStatus.close
                session.commit()
                logger.info(
                    'func=QuantfuryPositions.sync_closed_positions, balance_id={}, trade_id={}, close_price={}, '
                    'profit_and_loss_instrument={}'.format(
                        stored_position.trade.balance_id, stored_position.trade_id,
                        close_position['close_price'], close_position['profit_and_loss_instrument']))

    @classmethod
    def update_order(cls, order_response):
        order_id = order_response.get('orderId', None)
        try:
            cleaned_response = UtilsAPI.clean_response(order_response)
            cleaned_response['update_time'] = datetime.utcfromtimestamp(
                cleaned_response['update_time'] / 1000).strftime('%Y-%m-%d %H:%M:%S')
            cleaned_response['time'] = datetime.utcfromtimestamp(
                cleaned_response['time'] / 1000).strftime('%Y-%m-%d %H:%M:%S')

            session.query(QuantfuryOrders).filter_by(order_id=order_id).update(dict(**cleaned_response))
            session.commit()

        except Exception as e:
            order_response_str = ';'.join(
                ['{}:{}'.format(k, v) for k, v in cleaned_response.items() or order_response.items()])
            logger. error('func=Orders.update_order, order_id={}, order_response_str={}, msg={}'.format(
                order_id, order_response_str, e))

    @classmethod
    def cancel_limit_order(cls, client, position_id):
        order_id = None
        stage = 'API.cancel_order'  # for error control
        try:
            create_order_response = client.cancel(position_id)
            if 'data' not in create_order_response and create_order_response.get('code') != 'Success':
                raise ConnectionError('Empty Response')
            stage = 'SQL.update_order'
            order = session.query(QuantfuryPositions).filter_by(limit_id=position_id).first()
            order.status = QuantfuryOrderStatus.cancel
            session.commit()

            return order
        except Exception as e:
            logger.error('func=QuantfuryOrders.cancel_order, api_order_id={}, stage={}, msg={}'.format(
                order_id, stage, e))

            raise e

    @classmethod
    def rename_fields_from_response(cls, limit_response):
        new_response = {'limit_' + k: v for k, v in limit_response.items()}
        new_response['position_type'] = new_response.pop('limit_position_type')
        return new_response


class QuantfuryOrders(Base):
    __tablename__ = 'quantfury_orders'
    quantfury_order_id = Column(Integer, primary_key=True, nullable=False)
    api_order_id = Column(String, primary_key=True, nullable=False)
    api_position_id = Column(String, ForeignKey('quantfury_positions.api_position_id'), nullable=False)

    order_type = Column(Enum(QuantfuryOrderType), nullable=True)
    price = Column(Numeric, nullable=False)
    quantity = Column(Numeric, nullable=False)
    amount_instrument = Column(Numeric, nullable=False)
    amount_system = Column(Numeric, nullable=False)

    position = relationship("QuantfuryPositions", back_populates="quantfury_orders", uselist=False)


if __name__ == '__main__':
    from broker_clients import Balances, Trades

    balance_name = 'BTC_15_no_trend_1_no_break_even'
    balance = session.query(Balances).filter_by(name=balance_name).first()
    close_price, stop_loss, target_price = Utils.get_test_limit_stop_target_prices(
        SymbolMetadata.to_currency_pair_format(balance.symbol))


    trade = Trades(name=balance.name,
                   symbol=balance.symbol,
                   time_frame=balance.time_frame,
                   entry_date='2021-03-27 21:45:00.000000',
                   amount=10,
                   quantity=None,
                   stop_loss=stop_loss,
                   buy_price=close_price,
                   sell_price=None,
                   profit=None,
                   side='long',
                   low_moment_price=close_price,
                   balance_id=balance.balance_id)
    session.add(trade)
    session.commit()
    #login_result = Quantfury.login(balance.account.api_key, balance.account.secret_key)
    #client = Quantfury(login_result['data']['token'])
    client = Quantfury(balance.account.token)
    position = session.query(QuantfuryPositions).filter_by(api_position_id='52dd4715-961c-4a69-9693-bc0b07e65d22').first()
    #position.close_with_earnings(client, 49100)
    QuantfuryPositions().sync_closed_positions(client)
    QuantfuryPositions().sync_orders(client)
    order = QuantfuryPositions.register_order(
        client, trade_id=trade.trade_id, symbol=SymbolMetadata.to_currency_pair_format(balance.symbol),
        quote_order_quantity=trade.amount, buy_price=trade.buy_price, stop_price=trade.stop_loss,
        target_price=trade.sell_price, position_type=trade.side)

    print(order)

    QuantfuryPositions().cancel_limit_order(client, order.limit_id)



"""Create Quantfury orders model

Revision ID: 53cbb66861d9
Revises: e84250567e04
Create Date: 2021-07-31 11:10:57.991455

"""
from alembic import op
import sqlalchemy as sa
import enum


# revision identifiers, used by Alembic.
revision = '53cbb66861d9'
down_revision = 'e84250567e04'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    #op.execute("""
    #CREATE TYPE quantfuryorderstatus AS ENUM (
    #'LONG', 'SHORT');""")

    class QuantfuryOrderStatus(enum.Enum):
        hold = 'hold'
        open = 'open'
        close = 'close'
        cancel = 'cancel'
        target = 'target'

    class QuantfuryOrderType(enum.Enum):
        stop_order = 'STOP_ORDER'
        target_order = 'TARGET_ORDER'
        limit_order = 'LIMIT_ORDER'

    class QuantfuryPositionType(enum.Enum):
        long = 'long'
        short = 'short'

    op.create_table(
        'quantfury_positions',
        sa.Column('quantfury_order_id', sa.Integer, primary_key=True, nullable=False),
        sa.Column('trade_id', sa.Integer, nullable=False),
        sa.Column('limit_id', sa.String, nullable=False, unique=True),
        sa.Column('symbol', sa.String, nullable=False),

        sa.Column('limit_price', sa.Numeric, nullable=True),
        sa.Column('limit_quantity', sa.Numeric, nullable=False),
        sa.Column('limit_amount_instrument', sa.Numeric, nullable=False),
        sa.Column('limit_amount_system', sa.Numeric, nullable=False),
        sa.Column('limit_stop_order', sa.Numeric, nullable=False),
        sa.Column('limit_create_date', sa.DateTime, nullable=True),
        sa.Column('limit_execution_type', sa.Numeric, nullable=True),

        sa.Column('api_position_id', sa.String, nullable=True, unique=True),
        sa.Column('position_open_price', sa.Numeric, nullable=True),
        sa.Column('position_quantity', sa.Numeric, nullable=True),
        sa.Column('position_amount_instrument', sa.Numeric, nullable=True),
        sa.Column('position_amount_system', sa.Numeric, nullable=True),
        sa.Column('position_open_date', sa.DateTime, nullable=True),
        sa.Column('position_order', sa.Integer, nullable=True),
        sa.Column('position_session_currency', sa.String, nullable=True),
        sa.Column('position_session_id', sa.String, nullable=True),
        sa.Column('position_spread_adjustment_end_date', sa.DateTime, nullable=True),
        sa.Column('position_type', sa.Enum(QuantfuryPositionType), nullable=True),
        sa.Column('status', sa.Enum(QuantfuryOrderStatus), nullable=False, default=False),
        sa.Column('profit_and_loss_instrument', sa.Numeric, nullable=True),
        sa.Column('profit_and_loss_system', sa.Numeric, nullable=True),
        sa.Column('profit_and_loss_account', sa.Numeric, nullable=True),
        sa.Column('close_price', sa.Numeric, nullable=True),
        sa.Column('close_date', sa.DateTime, nullable=True)
    )

    op.create_foreign_key(
        constraint_name='quantfury_positions_trade_id_fkey',
        source_table='quantfury_positions',
        referent_table='trades',
        local_cols=['trade_id'],
        remote_cols=['trade_id'])

    op.add_column('symbol_metadata', sa.Column('currency_pair', sa.String(), nullable=True))
    
    op.create_table(
        'quantfury_orders',

        sa.Column('quantfury_order_id', sa.Integer, primary_key=True, nullable=False),
        sa.Column('api_order_id', sa.String, nullable=False),
        sa.Column('api_position_id', sa.String, nullable=False),

        sa.Column('order_type', sa.Enum(QuantfuryOrderType), nullable=False),
        sa.Column('price', sa.Numeric, nullable=True),
        sa.Column('quantity', sa.Numeric, nullable=False),
        sa.Column('amount_instrument', sa.Numeric, nullable=False),
        sa.Column('amount_system', sa.Numeric, nullable=False)
    )


    op.create_foreign_key(
        constraint_name='quantfury_orders_api_position_id_fkey',
        source_table='quantfury_orders',
        referent_table='quantfury_positions',
        local_cols=['api_position_id'],
        remote_cols=['api_position_id'])

    op.add_column('accounts', sa.Column('broker_name', sa.String(), nullable=True))
    op.execute("UPDATE accounts SET broker_name = 'binance'")
    op.alter_column('accounts', 'broker_name', nullable=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('quantfury_orders')
    op.drop_table('quantfury_positions')
    op.execute("DROP TYPE quantfuryorderstatus;")
    op.execute("DROP TYPE IF EXISTS quantfuryordertype;")
    op.execute("DROP TYPE IF EXISTS quantfurypositiontype;")
    # ### end Alembic commands ###



import decimal

from broker_clients.emulator_proxy import Broker
import logging
from broker_clients import session
from broker_clients import Balances, Trades

MINIMUM_TRADE_AMOUNT = 0
logger = logging.getLogger(__name__)


class BrokerEmulator(Broker):
    def __init__(self, symbol, time_frame, name, trade_in_trend):
        super().__init__(symbol, time_frame)
        self.name = name
        self.trade_in_trend = trade_in_trend
        self.balance_id = self.get_balance_id()
        if self.trading_budget() is None:
            self.set_trading_budget(100)

    @ property
    def stop_loss(self):
        trade = session.query(Trades).filter_by(name=self.name, sell_price=None, profit=None).first()
        if trade:
            return trade.stop_loss

    @ stop_loss.setter
    def stop_loss(self, value):
        trade = session.query(Trades).filter_by(name=self.name, sell_price=None, profit=None).first()
        if trade:
            trade.stop_loss = value
            session.commit()
            self._stop_loss = value

    @property
    def trade_id(self):
        trade = session.query(Trades).filter_by(name=self.name, sell_price=None, profit=None).first()
        if trade:
            return trade.trade_id
        return None

    def get_balance_id(self):
        balance = session.query(Balances).filter_by(name=self.name).first()
        if balance is None:
            raise ValueError('No Balance found with name: {}'.format(self.name))
        return balance.balance_id

    def trading_budget(self):
        balance = session.query(Balances).filter_by(name=self.name).first()
        if balance:
            return balance.free

    def set_trading_budget(self, value):
        if not self.trading_budget():

            balance = Balances(name=self.name, symbol=self._symbol, time_frame=self._time_frame, free=value, locked=0)
            session.add(balance)
        else:
            balance = session.query(Balances).filter_by(name=self.name).first()
            balance.free = value
        session.commit()

    def get_opened_trades(self):
        count = session.query(Trades).filter_by(name=self.name, sell_price=None, profit=None).count()
        return count

    def init_broker_object(self):
        count = session.query(Trades).filter_by(name=self.name, sell_price=None, profit=None).count()
        if count > 1:
            message = 'The balance {} has {} trades opened'.format(self.name, count)
            logger.error('func=init_broker_object, name={}, time_frame={}, symbol, msg={}'.format(
                self.name, self._time_frame, self._symbol, message))
            raise ValueError(message)

        trade = session.query(Trades).filter_by(name=self.name, sell_price=None, profit=None).first()
        if trade:
            self._price, self._stop_loss = trade.buy_price, trade.stop_loss
            self._amount, self.created_at = trade.amount, trade.created_at

    def market(self, side, amount, stop_loss=None, stop_loss_percentage=None, emulated_time=None):
        if side not in ['BUY', 'SELL'] or amount <= MINIMUM_TRADE_AMOUNT:
            raise ValueError('Invalid side: {} or amount: {} in market order'.format(side, amount))
        if stop_loss is None and stop_loss_percentage is None:
            raise ValueError('Stop loss must be set')
        if self.get_opened_trades() >= 1:
            raise ValueError('There are {} opened trades'.format(side))

        if side == 'BUY':
            self.price = decimal.Decimal(emulated_time['Close'])
            if not stop_loss:
                calculated_stop_loss = self.price - (self.price * stop_loss_percentage)
            self._amount = amount

            trade = Trades(name=self.name,
                           symbol=self._symbol,
                           time_frame=self._time_frame,
                           entry_date=emulated_time['date'],
                           amount=amount,
                           stop_loss=calculated_stop_loss,
                           buy_price=self.price,
                           sell_price=None,
                           profit=None,
                           side=side,
                           low_moment_price=emulated_time['Low'],
                           balance_id=self.balance_id)
            session.add(trade)
            session.commit()

    def take_profit(self, row=None):
        current_price = decimal.Decimal(row['Close'])
        delta = current_price - self.price
        profit_percentage = delta / current_price
        profit_amount = self._amount * profit_percentage
        self.set_sell_price(current_price, self._amount + profit_amount, row)
        self.clean_order()
        return {
            "profit_percentage": profit_percentage,
            "profit_amount": profit_amount
        }

    def set_sell_price(self, sell_price, profit, row):
        balance = session.query(Balances).filter_by(name=self.name).first()
        trade = session.query(Trades).filter_by(name=self.name, sell_price=None, profit=None).first()

        trade.sell_price = sell_price
        trade.profit = profit
        trade.exit_date = row['date']
        balance.free = profit
        session.commit()

    def has_open_order(self):
        return self.get_opened_trades() == 1

    def is_stop_loss_reached(self, prices):
        trade = session.query(Trades).filter_by(name=self.name, sell_price=None, profit=None).first()
        return trade.is_stop_loss_reached(prices)


if __name__ == '__main__':
    symbol, time_frame, name = 'BTCUSDT:BINANCE', '15', 'BTC'
    broker = BrokerEmulator(symbol, time_frame, name, None)
    side, amount, stop_loss, stop_loss_percentage, emulated_time = 'BUY', 100, 145, None, {
        "Close": 150, "date": "2021/01/14, 14:30:00"}
    broker.market(side, amount, stop_loss=stop_loss, emulated_time=emulated_time)



"""Add columns to balances to get dynamically the strategy approach

Revision ID: b1f321f39713
Revises: 7add5d2cc147
Create Date: 2021-03-05 13:15:17.642442

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy import MetaData, Table
import json


def choice_path(alternative_paths):
    import os
    alternative_paths = [i for i in alternative_paths if os.path.isfile(i)]
    if len(alternative_paths) == 0:
        raise FileExistsError('balances.json Does not exist')
    path = alternative_paths[0]
    return path


def get_module_path(module_name):
    parts = __file__.split('/')
    index = parts.index(module_name)
    module_path = '/'.join(parts[:index+1])
    return module_path


# revision identifiers, used by Alembic.
revision = 'b1f321f39713'
down_revision = '7add5d2cc147'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('balances', sa.Column('second_scale_from_database', sa.Boolean(), nullable=True))
    op.execute("UPDATE balances SET second_scale_from_database = True")
    op.alter_column('balances', 'second_scale_from_database', nullable=False)

    op.add_column('balances', sa.Column('trade_in_trend', sa.Boolean(), nullable=True))
    op.execute("UPDATE balances SET trade_in_trend = True")
    op.alter_column('balances', 'trade_in_trend', nullable=False)

    op.add_column('balances', sa.Column('stop_loss_percentage', sa.Numeric(), nullable=True))
    op.execute("UPDATE balances SET stop_loss_percentage = 0.3")
    op.alter_column('balances', 'stop_loss_percentage', nullable=False)

    op.add_column('balances', sa.Column('profit_percentage', sa.Numeric(), nullable=True))
    op.execute("UPDATE balances SET profit_percentage = 0.3")
    op.alter_column('balances', 'profit_percentage', nullable=False)
    module_path = get_module_path('broker_clients')
    alternative_paths = ['{}/data/balances.json'.format(module_path), '../data/balances.json']
    path = choice_path(alternative_paths)
    f = open(path)
    balances_initial_data = json.load(f)
    meta = MetaData(bind=op.get_bind())
    meta.reflect(only=('balances',))
    table = Table('balances', meta)
    op.bulk_insert(table, balances_initial_data)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    #f = open('data/balances.json')
    #balances_initial_data = json.load(f)
    #balances_names = [i['name'] for i in balances_initial_data]
    #records = session.query(Trades).filter(Trades.name.in_(balances_names))
    #if len(records.first()) > 0:
    #    records.delete(synchronize_session=False)
    #session.query(Balances).filter(Balances.name.in_(balances_names)).delete(synchronize_session='fetch')
    #session.commit()
    op.drop_column('balances', 'second_scale_from_database')
    op.drop_column('balances', 'trade_in_trend')
    op.drop_column('balances', 'stop_loss_percentage')
    op.drop_column('balances', 'profit_percentage')
    # ### end Alembic commands ###



"""Create Balances and Trades tables

Revision ID: 7d42c6c69b3a
Revises: 
Create Date: 2021-02-12 17:25:44.640224

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '7d42c6c69b3a'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'balances',
        sa.Column('balance_id', sa.Integer(), primary_key=True, autoincrement=True),
        sa.Column('name', sa.String(), nullable=False, unique=True),
        sa.Column('symbol', sa.String(), nullable=False),
        sa.Column('time_frame', sa.String(), nullable=False),
        sa.Column('free', sa.Numeric(), nullable=False),
        sa.Column('locked', sa.Numeric(), nullable=False),
        sa.Column('updated_at', sa.DateTime(), default=sa.func.now()),
        sa.Column('created_at', sa.DateTime(), default=sa.func.now()))

    op.create_table(
        'trades',
        sa.Column('trade_id', sa.Integer(), primary_key=True, autoincrement=True),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('symbol', sa.String(), nullable=False),
        sa.Column('time_frame', sa.String(), nullable=False),
        sa.Column('date', sa.DateTime(), nullable=False),
        sa.Column('amount', sa.Numeric(), nullable=False),
        sa.Column('stop_loss', sa.Numeric(), nullable=False),
        sa.Column('buy_price', sa.Numeric(), nullable=False),
        sa.Column('sell_price', sa.Numeric(), nullable=True),
        sa.Column('profit', sa.Numeric(), nullable=True),
        sa.Column('side', sa.String(), nullable=False),
        sa.Column('created_at', sa.DateTime(), default=sa.func.now()),
        sa.Column('updated_at', sa.DateTime(), default=sa.func.now()),
        sa.Column('balance_id', sa.Integer(), nullable=False))
    op.create_foreign_key(
        constraint_name='trades_balance_id_fkey',
        source_table='trades',
        referent_table='balances',
        local_cols=['balance_id'],
        remote_cols=['balance_id'])


def downgrade():
    op.drop_table('trades')
    op.drop_table('balances')



"""Add columns quantity, asset_commission and commission to Orders

Revision ID: 2aa4fb7b5553
Revises: 55199046e99a
Create Date: 2021-04-15 22:07:42.532328

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2aa4fb7b5553'
down_revision = '55199046e99a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('orders', sa.Column('binance_trade_id', sa.Integer(), nullable=True))
    op.add_column('orders', sa.Column('commission', sa.Numeric(), nullable=True))
    op.add_column('orders', sa.Column('commission_asset', sa.String(), nullable=True))
    op.alter_column('trades', 'amount',
               existing_type=sa.NUMERIC(),
               nullable=True)
    op.alter_column('trades', 'buy_price',
               existing_type=sa.NUMERIC(),
               nullable=True)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('trades', 'buy_price',
               existing_type=sa.NUMERIC(),
               nullable=False)
    op.alter_column('trades', 'amount',
               existing_type=sa.NUMERIC(),
               nullable=False)
    op.drop_column('orders', 'commission_asset')
    op.drop_column('orders', 'commission')
    op.drop_column('orders', 'binance_trade_id')
    # ### end Alembic commands ###



"""Add relationship between Accounts and Balances

Revision ID: c39e3caf08b6
Revises: b1f321f39713
Create Date: 2021-04-13 00:13:41.047697

"""
import os
from dotenv import load_dotenv
from alembic import op
import sqlalchemy as sa
from broker_clients import Accounts

# revision identifiers, used by Alembic.

revision = 'c39e3caf08b6'
down_revision = '14f9fd83af6a'
branch_labels = None
depends_on = None


load_dotenv()
cypher_key = os.getenv("CYPHER_KEY")

def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.create_table(
        'accounts',
        sa.Column('account_id', sa.Integer(), primary_key=True),
        sa.Column('name', sa.String(), unique=True, nullable=False),
        sa.Column('api_key', sa.String(), unique=True, nullable=False),
        sa.Column('secret_key', sa.String(), nullable=False),
        sa.Column('token', sa.String(), nullable=True),
        sa.Column('created_at', sa.DateTime(), default=sa.func.now()),
        sa.Column('updated_at', sa.DateTime(), default=sa.func.now())
    )

    op.bulk_insert(Accounts.__table__,
                   [
                       {
                           'name': 'testnet',
                           'api_key': 'u32stTD0rmEOaZteUJA9gx6ll2hyMhJNYoYjvp9y8eHPpQiGSebw54IayuRwXQAg',
                           'secret_key': 'zbht3ZFM2KT5uAKZHRiI5epvlhPEcGGXPNdPEArPSvPFxEHF9LkY2lwNtUVtvttU',
                       }
                   ])
    op.add_column('balances', sa.Column('account_id', sa.Integer(), nullable=True))
    op.execute("UPDATE balances SET account_id = 1")
    op.alter_column('balances', 'account_id', nullable=False)
    op.create_foreign_key(
        constraint_name='balances_account_id_fkey',
        source_table='balances',
        referent_table='accounts',
        local_cols=['account_id'],
        remote_cols=['account_id'])

    op.create_table(
        'orders',
        sa.Column('order_id', sa.Integer, primary_key=True, nullable=False),
        sa.Column('symbol', sa.String, nullable=False),
        sa.Column('order_list_id', sa.Integer, nullable=False),
        sa.Column('client_order_id', sa.String, nullable=False),
        sa.Column('transact_time', sa.DateTime, nullable=False),
        sa.Column('price', sa.Numeric, nullable=True),
        sa.Column('original_quantity', sa.Numeric, nullable=False),  # quantity requested
        sa.Column('executed_quantity', sa.Numeric, nullable=False),  # quantity filled
        sa.Column('cumulative_quote_quantity', sa.Numeric, nullable=False),

        sa.Column('status', sa.String, nullable=False),
        sa.Column('time_in_force', sa.String, nullable=False),
        sa.Column('type', sa.String, nullable=False),
        sa.Column('side', sa.String, nullable=False),

        sa.Column('stop_price', sa.Numeric, nullable=True),
        sa.Column('iceberg_quantity', sa.Numeric, nullable=True),  # For hiding the actual order quantity on big orders
        sa.Column('time', sa.DateTime, nullable=True),
        sa.Column('update_time', sa.DateTime, nullable=True),
        sa.Column('is_working', sa.Boolean, nullable=True),
        sa.Column('original_quote_order_quantity', sa.Numeric, nullable=True),  # quote quantity requested
    )

    op.create_table(
        'fills',
        sa.Column('fill_id', sa.Integer, primary_key=True, autoincrement=True),
        sa.Column('order_id', sa.Integer, nullable=False),
        sa.Column('price', sa.Numeric, nullable=False),
        sa.Column('quantity', sa.Numeric, nullable=False),
        sa.Column('commission', sa.Numeric, nullable=False),
        sa.Column('commission_asset', sa.String, nullable=False),
        sa.Column('trade_id', sa.Integer, nullable=False),
    )

    op.create_foreign_key(
        constraint_name='fills_order_id_fkey',
        source_table='fills',
        referent_table='orders',
        local_cols=['order_id'],
        remote_cols=['order_id'])

    op.create_table(
        'symbol_metadata',
        sa.Column('symbol_id', sa.Integer, primary_key=True, autoincrement=True),
        sa.Column('symbol', sa.String, nullable=False, unique=True),
        sa.Column('minimum_quantity_lot_size', sa.String, nullable=False),
        sa.Column('minimum_price_filter', sa.String, nullable=False),
        sa.Column('minimum_notional', sa.String, nullable=False)
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(constraint_name='fills_order_id_fkey', table_name='fills', type_='foreignkey')
    op.drop_table('fills')
    op.drop_table('orders')
    op.drop_constraint(constraint_name='balances_account_id_fkey', table_name='balances', type_='foreignkey')
    op.drop_column('balances', 'account_id')
    op.drop_table('accounts')


    op.drop_table('symbol_metadata')
    # ### end Alembic commands ###



from broker_clients import Accounts
from broker_clients.binance_proxy.broker.account import ExtendedAccount
from broker_clients.emulator_proxy.storage.models import SymbolMetadata
from broker_clients import Base, session
from broker_clients import Balances
a = Accounts(name='quantfury_test', api_key='5583198120', secret_key='Dando&perdiend0',
             token='z4Ti3RS6JwhnP7eShTo7ZSYk4s2PrfovzK7-EPAbhJ-Z5agupaICLdQoS-b8skEBpAEzcJHgaQbrS6I6nL7NcM27cZZNecXec6QmvsIW-UNwuFr_62sjsY4KpuPRPcVyF6kMWQJatcUw4240vx2cMtddyAj94i9X9mx_i6v5rYYWyvOJgzNFp8jRnEJmI9PzAM7vb8z9IKPNGGTznZBOwORwJ1heDdXKqlQT8WcoJ5YYtosBO6rWQ5dqjvKUye5QFSBteIFS-OSAOTZMCWTntmBCZxvdV81Jes1joaOZkqBfNjmnDVsvD5ZtOPJT8LgXqXAttkooAdYxUzhnDaxcIHvunznX8eYJx8nJTG_U6xLd4Kma7IOoICcYdF19lxP0orz0VmNykgbPbjdZjyxxBLzHJnOz2AP8M8YavWGyKrq9ewZEu2fAxlAuNUt_f7jfHp7mdyZ0Wp5u9dsIXQGPnG888KjmjDw11x2XRDU8RwCL-ti1', broker_name='quantfury')
session.add(a)
session.commit()

ea = ExtendedAccount('quantfury_test', 'BINANCE:BTCUSDT')
ea.update_symbol_metadata()

session.query(SymbolMetadata).filter_by(symbol='BTCUSDT').update({"currency_pair": "BTC/USDT"})
session.query(Balances).filter_by(name='BTC_15_no_trend_1_no_break_even').update({"account_id": 2})
session.commit()


import argparse
from broker_clients import Balances
from broker_clients import session


def parse_args():
    parser = argparse.ArgumentParser(prog='module')
    #parser.add_argument('--foo', action='store_true', help='foo help')
    subparsers = parser.add_subparsers(dest='module', help='sub-command help')

    parser_a = subparsers.add_parser('create_balance', help='Create a balance')
    parser_a.add_argument('--name', type=str, dest='name', required=True, help='Balance name')
    parser_a.add_argument('--symbol', type=str, dest='symbol', required=True, help='Balance symbol')
    parser_a.add_argument('--time_frame', type=str, dest='time_frame', required=True, help='Balance time frame')
    parser_a.add_argument('--free', type=float, dest='free', required=True, help='Balance budget')

    #parser_b = subparsers.add_parser('b', help='b help')
    #parser_b.add_argument('--baz', choices='XYZ', help='baz help')

    args = parser.parse_args()
    return args


if __name__ == '__main__':
    args = parse_args()
    if args.module == 'create_balance':
        balance = session.query(Balances).filter_by(name=args.name).first()
        if not balance:
            balance = Balances(name=args.name,
                               symbol=args.symbol,
                               time_frame=args.time_frame,
                               free=args.free,
                               locked=0)
            session.add(balance)
            session.commit()
            print(balance)
        else:
            print('Balance name already exists')




"""make nullable price column in orders

Revision ID: 5a0ada3a67ed
Revises: e96c68ba9fa5
Create Date: 2021-04-20 16:15:28.097624

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '5a0ada3a67ed'
down_revision = 'e96c68ba9fa5'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('orders', 'price',
               existing_type=sa.NUMERIC(),
               nullable=True)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('orders', 'price',
               existing_type=sa.NUMERIC(),
               nullable=False)
    # ### end Alembic commands ###



import re
from datetime import datetime

class UtilsAPI:
    @classmethod
    def to_snake_case(cls, phrase):
        return re.sub(r'(?<!^)(?=[A-Z])', '_', phrase).lower()

    @classmethod
    def convert_keys_to_snake_case(cls, json_response):
        json_response = cls.change_keys(json_response, cls.to_snake_case)
        return json_response

    @classmethod
    def rename_keys_with_prefix(cls, json_response, prefix):
        json_response = cls.change_keys(json_response, lambda s: prefix+s)
        return json_response

    @classmethod
    def change_keys(cls, obj, convert):
        if isinstance(obj, dict):
            new = {}
            for k, v in obj.items():
                new[convert(k)] = cls.change_keys(v, convert)
        elif isinstance(obj, (list, set, tuple)):
            new = [cls.change_keys(v, convert) for v in obj]
        else:
            return obj
        return new

    @classmethod
    def replace_abbreviations(cls, phrase):
        return phrase.replace('qty', 'quantity').replace('orig', 'original').replace('cummulative', 'cumulative')

    @classmethod
    def clean_response_field_names(cls, field_name):
        return cls.replace_abbreviations(cls.to_snake_case(field_name))

    @classmethod
    def clean_response(cls, response):
        if type(response) is dict:
            cleaned_response = {
                cls.clean_response_field_names(k): v if type(v) != list else [
                    cls.clean_response(i) for i in v] for k, v in response.items()}
            if 'transact_time' in cleaned_response:
                cleaned_response['transact_time'] = datetime.utcfromtimestamp(
                    cleaned_response['transact_time'] / 1000).strftime('%Y-%m-%d %H:%M:%S')

            if 'update_time' in cleaned_response:
                cleaned_response['update_time'] = datetime.utcfromtimestamp(
                    cleaned_response['update_time'] / 1000).strftime('%Y-%m-%d %H:%M:%S')

            if 'time' in cleaned_response:
                cleaned_response['time'] = datetime.utcfromtimestamp(
                    cleaned_response['time'] / 1000).strftime('%Y-%m-%d %H:%M:%S')

            return cleaned_response
        return response
